C
C     FFTE: A FAST FOURIER TRANSFORM PACKAGE
C
C     (C) COPYRIGHT SOFTWARE, 2000-2004, 2008-2014, 2020
C         ALL RIGHTS RESERVED
C                BY
C         DAISUKE TAKAHASHI
C         CENTER FOR COMPUTATIONAL SCIENCES
C         UNIVERSITY OF TSUKUBA
C         1-1-1 TENNODAI, TSUKUBA, IBARAKI 305-8577, JAPAN
C         E-MAIL: daisuke@cs.tsukuba.ac.jp
C
C
C     FACTORIZATION ROUTINE
C
C     FORTRAN77 SOURCE PROGRAM
C
C     WRITTEN BY DAISUKE TAKAHASHI
C
      SUBROUTINE FACTOR(N,IP)
      DIMENSION IP(*)
C
      IP(1)=0
      IP(2)=0
      IP(3)=0
      N2=N
      IF (MOD(N,2) .NE. 0 .AND. MOD(N,3) .NE. 0 .AND.
     1    MOD(N,5) .NE. 0) RETURN
   10 IF (N2 .LE. 1) RETURN
      IF (MOD(N2,2) .EQ. 0) THEN
        IP(1)=IP(1)+1
        N2=N2/2
        GO TO 10
      ELSE IF (MOD(N2,3) .EQ. 0) THEN
        IP(2)=IP(2)+1
        N2=N2/3
        GO TO 10
      ELSE IF (MOD(N2,5) .EQ. 0) THEN
        IP(3)=IP(3)+1
        N2=N2/5
        GO TO 10
      END IF
      RETURN
      END
      SUBROUTINE FACTOR2(N,KP)
      IMPLICIT REAL*8 (A-H,O-Z)
      DIMENSION IP(3),KP(*)
C
      CALL FACTOR(N,IP)
C
      DO 10 I=1,16
        KP(I)=0
   10 CONTINUE
      IF (N .EQ. 2) THEN
        KP(2)=1
        RETURN
      END IF
      H=0.0D0
      L16=IP(1)/4
      DO 110 J16=0,L16
        L15=MIN0(IP(2),IP(3))
        DO 100 J15=0,L15
          L12=MIN0((IP(1)-J16*4)/2,IP(2)-J15)
          DO 90 J12=0,L12
            L10=MIN0(IP(1)-J12*2-J16*4,IP(3)-J15)
            DO 80 J10=0,L10
              L9=(IP(2)-J12-J15)/2
              DO 70 J9=0,L9
                L8=MIN0((IP(1)-J10-J12*2-J16*4)/3,3)
                DO 60 J8=0,L8
                  L6=MIN0(IP(1)-J8*3-J10-J12*2-J16*4,IP(2)-J9*2-J12-J15)
                  DO 50 J6=0,L6
                    L5=IP(3)-J10-J15
                    DO 40 J5=0,L5
                      L4=MIN0((IP(1)-J6-J8*3-J10-J12*2-J16*4)/2,1)
                      DO 30 J4=0,L4
                        L3=MIN0(IP(2)-J6-J9*2-J12-J15,1)
                        DO 20 J3=0,L3
                          IF (IP(1) .EQ. J4*2+J6+J8*3+J10+J12*2+J16*4
     1                        .AND. IP(2) .EQ. J3+J6+J9*2+J12+J15 .AND.
     2                        IP(3) .EQ. J5+J10+J15) THEN
                            H2=HARMMEAN(J3,J4,J5,J6,J8,J9,J10,J12,J15,
     1                                  J16)
                            IF (H2 .GE. H) THEN
                              KP(3)=J3
                              KP(4)=J4
                              KP(5)=J5
                              KP(6)=J6
                              KP(8)=J8
                              KP(9)=J9
                              KP(10)=J10
                              KP(12)=J12
                              KP(15)=J15
                              KP(16)=J16
                              H=H2
                            END IF
                          END IF
   20                   CONTINUE
   30                 CONTINUE
   40               CONTINUE
   50             CONTINUE
   60           CONTINUE
   70         CONTINUE
   80       CONTINUE
   90     CONTINUE
  100   CONTINUE
  110 CONTINUE
      RETURN
      END
      FUNCTION HARMMEAN(J3,J4,J5,J6,J8,J9,J10,J12,J15,J16)
      IMPLICIT REAL*8 (A-H,O-Z)
      DIMENSION JP(16)
C
      DO 10 I=1,16
        JP(I)=0
   10 CONTINUE
      JP(3)=J3
      JP(4)=J4
      JP(5)=J5
      JP(6)=J6
      JP(8)=J8
      JP(9)=J9
      JP(10)=J10
      JP(12)=J12
      JP(15)=J15
      JP(16)=J16
      ICOUNT=0
      HARMMEAN=0.0D0
      DO 20 I=3,16
        IF (JP(I) .NE. 0) THEN
          HARMMEAN=HARMMEAN+DBLE(JP(I))/DBLE(I)
          ICOUNT=ICOUNT+JP(I)
        END IF
   20 CONTINUE
      HARMMEAN=DBLE(ICOUNT)/HARMMEAN
      RETURN
      END
      SUBROUTINE GETNXNY(N,NX,NY)
      IMPLICIT REAL*8 (A-H,O-Z)
      DIMENSION IP(3),LNX(3),LNY(3)
C
      ISQRTN=IDINT(DSQRT(DBLE(N)))
      CALL FACTOR(N,IP)
      DO 10 I=1,3
        LNX(I)=0
   10 CONTINUE
      IRES=ISQRTN
      DO 40 K=0,(IP(3)+1)/2
        DO 30 J=0,(IP(2)+1)/2
          DO 20 I=0,(IP(1)+1)/2
            NX=(2**I)*(3**J)*(5**K)
            IF (NX .LE. ISQRTN) THEN
              IRES2=ISQRTN-NX
              IF (IRES2 .LT. IRES) THEN
                LNX(1)=I
                LNX(2)=J
                LNX(3)=K
                IRES=IRES2
              END IF
            END IF
   20     CONTINUE
   30   CONTINUE
   40 CONTINUE
      DO 50 I=1,3
        LNY(I)=IP(I)-LNX(I)
   50 CONTINUE
      NX=(2**LNX(1))*(3**LNX(2))*(5**LNX(3))
      NY=(2**LNY(1))*(3**LNY(2))*(5**LNY(3))
      RETURN
      END
